import uuid


class Note(object):
    """
    A note.
    Currently has a title, tags, texr and a priority."""

    def __init__(self, title, tags, text, priority=1, unique_id=None):
        if unique_id is None:
            self.id = str(uuid.uuid4())
        else:
            self.id = unique_id
        self.title = title
        self.tags = self.__verify_tags(tags)
        self.text = text
        self.priority = priority

    def __str__(self):
        return str(self.to_dictionary)

    def __verify_tags(self, tags):
        return list(set(tags))

    def to_dictionary(self):
        return {
            self.id: {
                'title': self.title,
                'tags': self.tags,
                'text': self.text,
                'priority': self.priority,
            }
        }

    def increment_priority(self):
        self.priority += 1

    def decrement_priority(self):
        self.priority = 0 if self.priority - 1 < 0 else self.priority - 1

    def formatted_tags(self):
        return " // ".join(self.tags)

    def edit_note(self, title, tags, text):
        self.title = title
        self.tags = tags
        self.tags = self.__verify_tags(tags)
        self.text = text


class Tag(object):
    """A note tag, for categorisation/filtering"""

    def __init__(self, type, frequency):
        self.type = type
        self.frequency = frequency

    def __str__(self):
        return "[{0}] {1}".format(self.frequency, self.type)



import argparse

from .controller import Controller
from .interface import AerendeInterface
from .configuration import Configuration


def parse_args():
    parser = argparse.ArgumentParser(description='aerende')
    parser.add_argument('-c', '--config',
                        default='~/.andgeloman/aerende/config.yml',
                        help='The aerende config file')
    return parser.parse_args()


def main():
    try:
        args = parse_args()
        config = Configuration(args.config)
        interface = AerendeInterface()

        Controller(config=config, interface=interface)
    except KeyboardInterrupt:
        exit(0)


if __name__ == '__main__':
    main()



from os import path, makedirs
import yaml


class Configuration(object):

    DEFAULT_PALETTE = {
        'status_bar': ['black', 'white'],
        'edit_bar': ['black', 'light red'],
        'highlight_note': ['light blue', 'default'],
        'high_priority': ['light red', 'default']
    }

    DEFAULT_DATA_PATH = {
        'data_path': '~/.andgeloman/aerende/data.yml'
    }

    DEFAULT_KEY_BINDINGS = {
        'new_note': 'n',
        'delete_note': 'd',
        'edit_note': 'e',
        'increment_note_priority': '+',
        'decrement_note_priority': '-',
        'quit': 'q',
        'next_note': ['j', 'down'],
        'previous_note': ['k', 'up']
    }

    DEFAULT_CONFIG = {
        'palette': DEFAULT_PALETTE,
        'data_path': DEFAULT_DATA_PATH,
        'key_bindings': DEFAULT_KEY_BINDINGS
    }

    def __init__(self, configuration_path):
        self.__load_configuration(configuration_path)

    def __load_configuration(self, configuration_path):
        expanded_path = path.expanduser(configuration_path)
        if not path.exists(expanded_path):
            makedirs(path.dirname(expanded_path))
            with open(expanded_path, 'w') as config_file:
                yaml.dump(self.DEFAULT_CONFIG, config_file,
                          default_flow_style=False)
            self.palette = self.DEFAULT_PALETTE
            self.data_path = self.DEFAULT_DATA_PATH
            self.key_bindings = self.DEFAULT_KEY_BINDINGS
        else:
            self.__load_configuration_values(expanded_path)

    def __load_configuration_values(self, path):
        with open(path, 'r') as config_file:
            config_dict = yaml.load(config_file)
            if 'palette' in config_dict:
                self.palette = self.__validate_configuration(
                    'palette', config_dict, self.DEFAULT_PALETTE)
            else:
                self.palette = self.DEFAULT_PALETTE

            if 'data_path' in config_dict:
                self.data_path = self.__validate_configuration(
                    'data_path', config_dict, self.DEFAULT_DATA_PATH)
            else:
                self.data_path = self.DEFAULT_DATA_PATH

            if 'key_bindings' in config_dict:
                self.key_bindings = self.__validate_configuration(
                    'key_bindings', config_dict, self.DEFAULT_KEY_BINDINGS)
            else:
                self.key_bindings = self.DEFAULT_KEY_BINDINGS

    def __validate_configuration(self, config_key, config_dict, defaults):
        config_item = {}
        for key, default_value in defaults.items():
            if key in config_dict[config_key]:
                config_item[key] = config_dict[config_key][key]
            else:
                config_item[key] = default_value
        return config_item

    def get_palette(self):
        return [[k] + v for k, v in self.palette.items()]

    def get_data_path(self):
        return self.data_path['data_path']

    def get_key_bindings(self):
        return self.key_bindings



from urwid import (Columns,
                   LineBox,
                   Text,
                   Frame,
                   ListBox,
                   AttrMap,
                   Padding,
                   SimpleListWalker,
                   signals,
                   connect_signal,
                   emit_signal,
                   Edit,
                   WidgetWrap)
from functools import reduce
import sys
import tempfile
import os
from subprocess import call

from . import version, title


class AerendeInterface(Columns):
    """
    Creates the interface for Aerende, as well as providing refresh and drawing
    functions.
    """

    def __init__(self):
        self.notes_frame = NotesFrame([])
        self.tags_list = TagsListBox([])
        Columns.__init__(self, [(18, self.tags_list), self.notes_frame])

    # [ Notes ]

    def draw_notes(self, notes):
        self.contents[1][0].draw_notes(notes)

    # [ Tags ]

    def draw_tags(self, tags):
        tag_list = self.contents[0][0]
        tag_list.draw_tags(tags)
        new_options = self.options('given', tag_list.get_max_width(tags),
                                   False)
        self.contents[0] = (tag_list, new_options)

    def show_note_editor(self, done_handler, note_to_edit=None):
        self.notes_frame.show_note_editor(done_handler, note_to_edit)

    def get_note_editor(self):
        return self.notes_frame.editor

    def focus_next_note(self):
        self.notes_frame.focus_next_note()

    def focus_previous_note(self):
        self.notes_frame.focus_previous_note()

    def focus_first_note(self):
        self.notes_frame.focus_first_note()

    def get_focused_note(self):
        return self.notes_frame.get_focused_note()


class NoteWidget(LineBox):
    """Widget for displaying a note"""

    def __init__(self, note):
        self.note = note

        header = self._create_header(note)
        footer = self._create_footer(note)
        note_display = Text("{0}\n\n{1}\n\n{2}".format(
            header, note.text, footer))
        LineBox.__init__(self, note_display)

    def _create_header(self, note):
        return "{0} :: {1}".format(note.priority, note.title)

    def _create_footer(self, note):
        footer_template = "[ {0} ]"
        return footer_template.format(note.formatted_tags())

    def selectable(self):
        return True

    def is_high_priority(self):
        return self.note.priority >= 10


class NotesFrame(Frame):
    """"Frame for displaying notes and status"""

    def __init__(self, notes):
        note_widgets = self._create_note_widgets(notes)
        self.notes = NotesListBox(note_widgets)
        self.notes.focus_first()
        self.status = self._create_statusbar(notes)
        Frame.__init__(self, self.notes, footer=self.status)

    def _create_note_widgets(self, notes):
        note_widgets = []
        for note in notes:
            note_widgets.append(AttrMap(NoteWidget(note),
                                        None,
                                        'highlight_note'))
        return note_widgets

    def _create_statusbar(self, notes):
        aerende_text = Text("{0} :: {1}".format(title, version))
        notes_text = Text(" [ {0} ]".format(len(notes)), align='right')
        columns = Columns([aerende_text, notes_text])
        return AttrMap(Padding(columns, left=1, right=1), 'status_bar')

    def draw_notes(self, notes):
        notes.sort(key=lambda note: note.priority, reverse=True)
        note_widgets = self._create_note_widgets(notes)
        self.body.refresh_notes(note_widgets)
        self.footer = self._create_statusbar(notes)
        self.set_footer(self.footer)

    def show_note_editor(self, done_handler, note_to_edit=None):
        self.editor = AttrMap(NoteEditor(done_handler, note_to_edit),
                              'edit_bar')
        self.footer = self.editor
        self.set_footer(self.footer)
        self.set_focus('footer')

    def focus_next_note(self):
        self.body.focus_next()

    def focus_previous_note(self):
        self.body.focus_previous()

    def focus_first_note(self):
        self.body.focus_first()

    def get_focused_note(self):
        note_widget, _ = self.body.get_focus()
        return note_widget.original_widget.note


class TagsListBox(ListBox):
    """ListBox widget for displaying a list of tags"""

    def __init__(self, tags):
        tag_widgets = self._create_tag_widgets(tags)
        self.list_content = SimpleListWalker(tag_widgets)
        ListBox.__init__(self, self.list_content)

    def _create_tag_widgets(self, tags):
        tag_widgets = []
        for tag in tags:
            tag_widgets.append(Text(str(tag)))
        return tag_widgets

    def draw_tags(self, tags):
        tags.sort(key=lambda tag: tag.frequency, reverse=True)
        tag_widgets = self._create_tag_widgets(tags)
        self.list_content[:] = tag_widgets

    def get_max_width(self, tags):
        max_width = 30
        min_width = 18
        width = len(str(reduce((lambda x, y:
                                x if len(str(x)) > len(str(y)) else y),
                               tags))) + 2
        return max(min(max_width, width), min_width)


class NoteEditor(WidgetWrap):

    __metaclass__ = signals.MetaSignals
    signals = ['done']

    def __init__(self, done_handler, note=None):
        self.modes = ['title', 'tags', 'text']
        self.mode = self.modes[0]
        self.note = note
        if note is None:
            self.title = ''
            self.tags = ''
            self.text = ''
        else:
            self.title = note.title
            self.tags = note.formatted_tags()
            self.text = note.text

        self.editor = Edit(u'title :: ', self.title)
        connect_signal(self, 'done', done_handler)
        WidgetWrap.__init__(self, self.editor)

    def keypress(self, size, key):
        if key == 'enter':
            if self.mode == 'title':
                self.title = self.editor.get_edit_text()
                self.init_tags_mode()
            elif self.mode == 'tags':
                self.tags = self.editor.get_edit_text()
                self.init_text_mode()

        elif key == 'esc':
            self.emit_done()
            return

        size = size,
        self.editor.keypress(size, key)

    def init_tags_mode(self):
        self.mode = self.modes[1]
        self.editor.set_caption('tags :: ')
        self.editor.set_edit_text(self.tags)

    def init_text_mode(self):
        self.mode = self.modes[2]
        editor = os.environ.get('EDITOR', 'vim')
        with tempfile.NamedTemporaryFile(prefix="aerende_tmp",
                                         suffix=".tmp") as temp:
            temp.write(self.text.encode('utf-8'))
            temp.flush()
            call([editor, temp.name])
            temp.seek(0)
            self.text = temp.read().decode('utf-8')
            os.system('clear')
            self.emit_done((self.title, self.tags, self.text))

    def emit_done(self, note=None):
        emit_signal(self, 'done', note, self.note)


class NotesListBox(ListBox):

    def __init__(self, contents):
        self.list_walker = SimpleListWalker(contents)
        ListBox.__init__(self, self.list_walker)

    def refresh_notes(self, notes):
        self.list_walker[:] = notes
        self.set_selected_style()

    def focus_next(self):
        _, position = self.get_focus()
        if position is None:
            return

        if position + 1 >= len(self.body):
            self.set_focus(position)
        else:
            self.set_focus(position + 1)
        self.set_selected_style()

    def focus_previous(self):
        _, position = self.get_focus()
        if position is None:
            return

        if position - 1 < 0:
            self.set_focus(position)
        else:
            self.set_focus(position - 1)
        self.set_selected_style()

    def focus_first(self):
        if len(self.body):
            self.set_focus(0)
            self.set_selected_style()

    def set_selected_style(self):
        focused, position = self.get_focus()
        if focused:
            focused.set_attr_map({None: 'highlight_note'})
        for object in self.body:
            if object is not focused:
                if object.original_widget.is_high_priority():
                    object.set_attr_map({None: 'high_priority'})
                else:
                    object.set_attr_map({None: None})



from pathlib import Path
import yaml
from os import path
import uuid
from urwid import MainLoop, ExitMainLoop
from functools import reduce

from .models import Note, Tag


class KeyHandler(object):

    def __init__(self, controller, config):
        self.controller = controller
        self.config = config
        self.editor = False

    def handle(self, input, *args, **kwargs):
        if self.is_keyboard_input(input):
            key = ''.join(input)
            self.handle_key(key)

    def is_key_bound(self, key, name):
        try:
            bound_keys = self.config.get_key_bindings()[name]
        except KeyError:
            return False
        else:
            return key in bound_keys

    def is_keyboard_input(self, input):
        if input:
            return reduce(lambda x, y: x and y,
                          map(lambda s: isinstance(s, str), input))

    def handle_key(self, key):
        if self.controller.editor_mode:
            size = 20
            self.editor.keypress(size, key)
            return

        if not self.controller.editor_mode:
            if self.is_key_bound(key, 'new_note'):
                self.controller.show_note_editor(
                    self.controller.edit_note_handler)
            elif self.is_key_bound(key, 'delete_note'):
                self.controller.delete_focused_note()
            elif self.is_key_bound(key, 'edit_note'):
                self.controller.show_note_editor(
                    self.controller.edit_note_handler, True)
            elif self.is_key_bound(key, 'increment_note_priority'):
                self.controller.increment_focused_note_priority()
            elif self.is_key_bound(key, 'decrement_note_priority'):
                self.controller.decrement_focused_note_priority()
            elif self.is_key_bound(key, 'quit'):
                self.controller.exit()
            elif self.is_key_bound(key, 'next_note'):
                self.controller.focus_next_note()
            elif self.is_key_bound(key, 'previous_note'):
                self.controller.focus_previous_note()


class Controller(object):

    def __init__(self, config, interface):
        self.config = config
        self.data_path = path.expanduser(self.config.get_data_path())
        self.notes = self.load_notes()
        self.tags = self.load_tags(self.notes)
        self.interface = interface
        self.editor_mode = False

        self.key_handler = KeyHandler(self, config)
        self.loop = MainLoop(interface,
                             config.get_palette(),
                             input_filter=self.key_handler.handle)
        self.refresh_interface()
        self.interface.focus_first_note()
        self.loop.run()

    def load_notes(self):
        # If no notes, create empty file
        if Path(self.data_path).is_file():
            with open(self.data_path, 'r') as data_file:
                note_yaml = yaml.load(data_file)
                notes = []

                if note_yaml is None:
                    return notes

                for unique_id, note in note_yaml.items():
                    notes.append(Note(note['title'],
                                      note['tags'],
                                      note['text'],
                                      note['priority'],
                                      unique_id))
                return notes
        else:
            open(self.data_path, 'x')
            return []

    def write_notes(self):
        with open(self.data_path, 'w') as data_file:
            for note in self.notes:
                yaml.dump(note.to_dictionary(),
                          data_file,
                          default_flow_style=False)

    def create_note(self, title, tags, text):
        note = Note(title, tags, text)
        self.notes.append(note)

    def delete_note(self, unique_id):
        for index, note in enumerate(self.notes):
            if note.id == unique_id:
                del self.notes[index]
                break

    def update_note(self, new_note):
        for index, note in enumerate(self.notes):
            if note.id == new_note.id:
                note = new_note
                break

    def delete_focused_note(self):
        note = self.interface.get_focused_note()
        self.delete_note(note.id)
        self.write_notes()

        self.refresh_interface()

    def load_tags(self, notes):
        tags = {'ALL': 0}
        tag_widgets = []
        for note in notes:
            tags['ALL'] += 1
            for tag in note.tags:
                if tag in tags:
                    tags[tag] += 1
                else:
                    tags[tag] = 1
        for tag in tags:
            tag_widgets.append(Tag(tag, tags[tag]))
        return tag_widgets

    def increment_focused_note_priority(self):
        note = self.interface.get_focused_note()
        note.increment_priority()
        self.write_notes()
        self.refresh_interface()

    def decrement_focused_note_priority(self):
        note = self.interface.get_focused_note()
        note.decrement_priority()
        self.write_notes()
        self.refresh_interface()

    def refresh_interface(self):
        self.interface.draw_notes(self.notes)
        self.tags = self.load_tags(self.notes)
        self.interface.draw_tags(self.tags)

    def show_note_editor(self, note_handler, edit_focused_note=False):
        note_to_edit = None
        if edit_focused_note:
            note_to_edit = self.interface.get_focused_note()
        self.editor_mode = True
        self.interface.show_note_editor(note_handler, note_to_edit)
        self.key_handler.editor = self.interface.get_note_editor()

    def edit_note_handler(self, note, original_note=None):
        if note is not None:
            title = note[0]
            tags = self._convert_tag_input(note[1])
            text = note[2]
            if original_note is not None:
                original_note.edit_note(title, tags, text)
                self.update_note(original_note)
            else:
                self.create_note(title, tags, text)
            self.write_notes()

            # Restart the loop.. Seems to work?
            self.loop.stop()
            self.loop.start()

        self.refresh_interface()
        self.editor_mode = False

    def _convert_tag_input(self, tag_text):
        split_tags = tag_text.split('//')
        return list(map(lambda tag: tag.strip(), split_tags))

    def exit(self):
        raise ExitMainLoop()

    def focus_next_note(self):
        self.interface.focus_next_note()

    def focus_previous_note(self):
        self.interface.focus_previous_note()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

setup(
    name='aerende',
    version='0.1.1',
    author='Mika Naylor (Autophagy)',
    author_email='mail@autophagy.io',
    url='https://github.com/Autophagy/aerende',
    description='A post-it note/reminders tool',
    long_description='''Ærende is a small, python based note taking application.
        Works offline and in the terminal via a curses interface.
        Navigation via vim-esque keys.
        Designed to slot easily into my comm workspace alongside weechat and neomutt.''',
    entry_points={
        'console_scripts': [
            'aerende = aerende.__main__:main',
        ]
    },
    packages=['aerende'],
    install_requires=[
        'PyYAML',
        'urwid'
    ],
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
)
